import { __rest } from "tslib";
import { request as httpRequest } from "http";
import { request as httpsRequest } from "https";
import db from "mime-db";
import { URL } from "url";
import { identity as identityEnumerable, lift, emptyStream, sink, } from "@reactive-js/async-enumerable";
import { createDisposableValue, AbstractDisposable, } from "@reactive-js/disposable";
import { createRedirectHttpRequest, mediaTypeIsCompressible, parseHttpResponseFromHeaders, httpRequestToUntypedHeaders, } from "@reactive-js/http";
import { createBufferStreamSinkFromWritable, transform, createBufferStreamFromReadable, } from "@reactive-js/node";
import { createObservable, ofValue, concatMap, onNotify, scan, subscribe, } from "@reactive-js/observable";
import { isSome, none } from "@reactive-js/option";
import { pipe, compose } from "@reactive-js/pipe";
import { supportedEncodings, createEncodingCompressTransform, getFirstSupportedEncoding, } from "./httpContentEncoding";
import { decodeHttpContentResponse } from "./httpResponse";
const spyScanner = ([uploaded, total], ev) => ev.type === 1
    ? [ev.data.length, total + uploaded]
    : [-1, total + uploaded];
const send = (request, requestOptions) => {
    const { contentEncoding } = requestOptions, nodeOptions = __rest(requestOptions, ["contentEncoding"]);
    const { content } = request;
    const newContent = isSome(content) && isSome(contentEncoding)
        ? Object.assign(Object.assign({}, content), { contentEncodings: [...content.contentEncodings, contentEncoding] }) : content;
    const requestWithAcceptEncodingsAndContentEncoding = Object.assign(Object.assign({}, request), { acceptedEncodings: supportedEncodings, content: newContent });
    const { method, uri } = requestWithAcceptEncodingsAndContentEncoding;
    const headers = httpRequestToUntypedHeaders(requestWithAcceptEncodingsAndContentEncoding);
    const nodeRequestUrl = uri instanceof URL ? uri : new URL(uri.toString());
    const nodeRequestOptions = Object.assign(Object.assign({}, nodeOptions), { headers,
        method });
    return uri.protocol === "https:"
        ? httpsRequest(nodeRequestUrl, nodeRequestOptions)
        : uri.protocol === "http:"
            ? httpRequest(nodeRequestUrl, nodeRequestOptions)
            : (() => {
                throw new Error();
            })();
};
const createOnSubscribe = (request, requestOptions, encodeContent, decodeHttpContentResponse) => (subscriber) => {
    var _a, _b;
    subscriber.dispatch({ type: 1, request });
    const req = send(request, requestOptions);
    const onError = (cause) => {
        subscriber.dispose({ cause });
    };
    req.on("error", onError);
    const onResponse = (msg) => {
        var _a;
        subscriber.add(() => msg.destroy());
        const response = pipe(parseHttpResponseFromHeaders((_a = msg.statusCode) !== null && _a !== void 0 ? _a : -1, msg.headers, createBufferStreamFromReadable(() => msg)), decodeHttpContentResponse);
        const disposableResponse = createDisposableValue(response, _ => {
            msg.destroy();
        });
        subscriber.dispatch({
            type: 4,
            request,
            response: disposableResponse,
        });
    };
    req.on("response", onResponse);
    const reqBody = createBufferStreamSinkFromWritable(() => req, false);
    const doOnNotify = ([count, total]) => {
        const ev = count < 0
            ? { type: 3, request }
            : { type: 2, request, total };
        subscriber.dispatch(ev);
    };
    const spyEnumerator = pipe(identityEnumerable(), lift(compose(scan(spyScanner, () => [0, 0]), onNotify(doOnNotify)))).enumerateAsync(subscriber);
    const contentBody = pipe((_b = (_a = request.content) === null || _a === void 0 ? void 0 : _a.body) !== null && _b !== void 0 ? _b : emptyStream(), lift(onNotify(ev => spyEnumerator.dispatch(ev))), encodeContent);
    subscriber.add(spyEnumerator).add(_ => {
        req.abort();
        req.removeAllListeners();
        req.destroy();
    });
    const onContinue = () => {
        subscriber.add(pipe(sink(contentBody, reqBody), subscribe(subscriber)));
    };
    if (request.expectContinue) {
        req.on("continue", onContinue);
    }
    else {
        onContinue();
    }
};
const identity = (x) => x;
const requestIsCompressible = (request) => {
    const { content } = request;
    return isSome(content)
        ? mediaTypeIsCompressible(db)(content.contentType)
        : false;
};
class HttpClientImpl extends AbstractDisposable {
    constructor(clientOptions) {
        super();
        this.clientOptions = clientOptions;
    }
    send(request, options = {}) {
        const _a = this.clientOptions, { agent, insecureHTTPParser, maxHeaderSize, shouldEncode: shouldEncodeOption } = _a, zlibOptions = __rest(_a, ["agent", "insecureHTTPParser", "maxHeaderSize", "shouldEncode"]);
        const { acceptedEncodings = supportedEncodings } = options;
        const contentEncoding = getFirstSupportedEncoding(acceptedEncodings);
        const shouldEncodeOptionResult = isSome(shouldEncodeOption)
            ? shouldEncodeOption(request)
            : none;
        const shouldEncode = isSome(shouldEncodeOptionResult)
            ? shouldEncodeOptionResult
            : requestIsCompressible(request);
        const contentEncoder = isSome(contentEncoding) && shouldEncode
            ? transform(createEncodingCompressTransform(contentEncoding, zlibOptions))
            : identity;
        const decodeHttpContentResponseWithOption = decodeHttpContentResponse(zlibOptions);
        const requestOptions = {
            agent,
            contentEncoding,
            insecureHTTPParser,
            maxHeaderSize,
        };
        return createObservable(createOnSubscribe(request, requestOptions, contentEncoder, decodeHttpContentResponseWithOption));
    }
}
export const creatHttpClient = (clientOptions = {}) => new HttpClientImpl(clientOptions);
const redirectCodes = [
    301,
    302,
    303,
    307,
    308,
];
export const createDefaultHttpResponseHandler = (httpClient, maxRedirects = 10) => {
    const handleResponse = (status) => {
        var _a;
        if (status.type === 4) {
            const { request, response } = status;
            const { location, preferences, statusCode } = response.value;
            const acceptedEncodings = (_a = preferences === null || preferences === void 0 ? void 0 : preferences.acceptedEncodings) !== null && _a !== void 0 ? _a : [];
            const shouldRedirect = redirectCodes.includes(statusCode) &&
                isSome(location) &&
                maxRedirects > 0;
            const [newRequest, newAcceptedEncodings] = shouldRedirect
                ? [
                    createRedirectHttpRequest(response.value)(request),
                ]
                : statusCode === 417
                    ? [Object.assign(Object.assign({}, request), { expectContinue: false })]
                    : statusCode === 415 &&
                        acceptedEncodings.length > 0
                        ? [request, acceptedEncodings]
                        : [request];
            if (request !== newRequest || isSome(newAcceptedEncodings)) {
                response.dispose();
                return pipe(httpClient.send(newRequest, {
                    acceptedEncodings: newAcceptedEncodings,
                }), createDefaultHttpResponseHandler(httpClient, maxRedirects - 1));
            }
        }
        return ofValue(status);
    };
    return concatMap(handleResponse);
};
//# sourceMappingURL=httpClient.js.map